import bpy
import bpy.utils.previews
import os
from . import views
from . import models
from . import properties
from . import utils
from . import strings

class Main(properties.IMocopiAvatarPropertyListener, views.IAvatarPanelListener):

    classes = [
        properties.MocopiAvatarProperty,
        properties.MocopiProperty,
        views.MOCOPI_RECEIVER_PT_Panel, 
        views.MOCOPI_RECEIVER_OT_Panel,
        views.MOCOPI_RECEIVER_OT_AvatarPanel,
    ]
    image_collection = None

    avatars: list[models.Avatar] = []
    max_num: int = 1

    def __init__(self):
        pass

    # Public

    def register(self):

        # 多言語対応
        bpy.app.translations.register(__name__, strings.locals)

        # 画像リソース
        Main.image_collection = bpy.utils.previews.new()
        Main.image_collection.load("mocopi_icon", f"{os.path.dirname(__file__)}/assets/mocopiSDK_Color.png", 'IMAGE')

        # リスナー登録
        properties.MocopiAvatarProperty.listener = self
        views.MOCOPI_RECEIVER_PT_Panel.listener = self
        views.MOCOPI_RECEIVER_PT_AvatarPanel.listener = self

        # パネルとオペレータ
        for c in Main.classes:
            bpy.utils.register_class(c)

        # プロパティ
        properties.MocopiProperty.register()

        # # 初回パネルの追加
        # bpy.app.timers.register(lambda: self.__add_avatar(bpy.context), first_interval=1)

        # 定期処理
        bpy.app.timers.register(self.__update, persistent=True)

    def unregister(self):

        # # 定期処理
        # bpy.app.timers.unregister(self.__update)

        # プロパティ
        properties.MocopiProperty.unregister()

        # パネルとオペレータ
        for c in Main.classes:
            bpy.utils.unregister_class(c)
        
        # 画像リソース
        bpy.utils.previews.remove(Main.image_collection)

        # 多言語対応
        bpy.app.translations.unregister(__name__)

    # IMocopiAvatarPropertyListener

    def on_rig_updated(self, context: bpy.types.Context, id: int, rig: bpy.types.Object):

        avatar = self.find_avatar(id)
        if avatar is None:
            return 
        
        # オートリターゲット
        prop: properties.MocopiAvatarProperty = utils.find_prop(id)
        prop.auto_retarget(rig)

        avatar.set_rig(rig)
        avatar.retarget()

    def on_bone_updated(self, context: bpy.types.Context, id: int, rig: str):
        avatar = self.find_avatar(id)
        if avatar is None:
            return
        avatar.retarget()

    # IPanelListener

    def on_add_button_clicked(self, op: views.MOCOPI_RECEIVER_OT_Panel, context: bpy.types.Context):

        id = Main.max_num
        Main.max_num += 1

        panel_class = self.__add_panel(id, context)

        # アバター
        avatar = models.Avatar(id, panel_class)
        self.avatars.append(avatar)

    # IAvatarPanelListener

    def on_drawed(self, panel: views.MOCOPI_RECEIVER_PT_AvatarPanel, context: bpy.types.Context, id: int) -> models.Avatar:

        prop: properties.MocopiAvatarProperty = utils.find_prop(id)
        if prop is None:
            return None
        
        avatar = self.find_avatar(id)
        if avatar is None:
            return
        
        if avatar.rig and not utils.is_valid(avatar.rig):
            if avatar.running:
                avatar.stop()
            avatar.set_rig(None)

        avatar.prop = prop

        return avatar

    def on_remove_button_clicked(self, op: views.MOCOPI_RECEIVER_OT_AvatarPanel, context: bpy.types.Context, id: int):

        avatar = self.find_avatar(id)
        if avatar is None:
            return

        self.__remove_panel(id, context)

        # アバター
        self.avatars.remove(avatar) 
        del avatar

    def on_connect_button_clicked(self, op: views.MOCOPI_RECEIVER_OT_AvatarPanel, context: bpy.types.Context, id: int):

        avatar = self.find_avatar(id)
        if not avatar.running:

            if not avatar.has_rig():
                op.report({'WARNING'}, strings.get('msg_target_is_not_armature'))
                return
            
            # mocopi接続
            avatar.retarget()
            avatar.run()

            if not avatar.running:
                op.report({'WARNING'}, strings.get('msg_confirm_port'))
                return

            # 記録開始
            if not bpy.context.screen.is_animation_playing:
                bpy.ops.screen.animation_play() # キーフレーム

        else: 

            # 記録終了
            if bpy.context.screen.is_animation_playing:
                bpy.ops.screen.animation_play() # キーフレーム

            # mocopi切断            
            avatar.stop()

    def on_reboot_button_clicked(self, op, context: bpy.types.Context, id: int):
        self.__reboot_panel(id, context)

    def on_debug_button_clicked(self, op: views.MOCOPI_RECEIVER_OT_AvatarPanel, context: bpy.types.Context, id: int):
        avatar = self.find_avatar(id)
        avatar.debug_id += 1
        # avatar.location0 = None
        print(f"debug_id: {avatar.debug_id}")

    # Private

    def __update(self):
    # def __update(self, scene, depsgraph):
        for avatar in self.avatars:
            avatar.update()
        return 0

    def __add_panel(self, id: int, context: bpy.types.Context) -> type:

        # プロパティ
        prop = utils.find_prop(id)
        if prop is None:
            prop: properties.MocopiAvatarProperty = context.scene.mocopi_property.avatars.add() 
            prop.init()
            prop.id = id

        # パネル
        panel_class = type(
            'MOCOPI_RECEIVER_AvatarPanel_' + str(id),
            (views.MOCOPI_RECEIVER_PT_AvatarPanel,),
            {
                'bl_idname': 'MOCOPI_RECEIVER_PT_AvatarPanel_' + str(id),
                'bl_label': f"{strings.get('mocopi_receiver')} [{id}]",
                'id': id,
            }
        )

        Main.classes.append(panel_class)
        bpy.utils.register_class(panel_class)

        return panel_class
    
    def __reboot_panel(self, id: int, context: bpy.types.Context):

        avatar = self.find_avatar(id)
        if avatar is None:
            return
        
        self.__remove_panel(id, context)
        # id = Main.max_num
        # Main.max_num += 1
        panel_class = self.__add_panel(id, context)
        
        avatar.id = id
        avatar.panel_class = panel_class

        prop = utils.find_prop(id)
        prop.rig = next((o for o in bpy.data.objects if o.name == avatar.rig_name), None)
    
    def __remove_panel(self, id: int, context: bpy.types.Context):

        avatar = self.find_avatar(id)
        if avatar.running:
            avatar.stop()

        # パネル
        Main.classes.remove(avatar.panel_class)
        bpy.utils.unregister_class(avatar.panel_class)

        # 記録終了
        if len(list(filter(lambda a: a.running, Main.avatars))) == 0:
            bpy.ops.screen.animation_cancel() # キーフレーム

    def find_avatar(self, id: int) -> models.Avatar:
        for avatar in self.avatars:
            if avatar.id == id:
                return avatar
        return None