import bpy
import threading
import asyncio
import socket
import struct
import math
from typing import List
from mathutils import Vector, Quaternion, Matrix
from . import properties
from . import utils

class BoneData:

    def __init__(self, bone_id: int):
        self.bone_id = bone_id
        self.bone: bpy.types.PoseBone = None

class Avatar:

    def __init__(self, id: int, panel_class: type):

        self.id = id
        self.panel_class = panel_class

        self.rig: bpy.types.Object = None
        self.rig_name: str = ''
        self.prop: properties.MocopiAvatarProperty = None
        self.action: bpy.types.Action = None

        self.running: bool = False
        self.udp_socket: socket.socket = None
        self.thread: threading.Thread = None

        self.last_data: MocopiData = None
        self.location0: Vector = None
        self.updating: bool = False
        self.rig_scale: float = 1.0

        self.debug_id: int = 0

        self.bone_data_list: List[BoneData] = []
        for bone_id in range(MocopiData.MAX_BONE): 
            self.bone_data_list.append(BoneData(bone_id))

    # Public

    def update(self):

        if not self.has_rig() or not self.running or not self.last_data:
            self.updating = False
            return
        
        self.updating = True

        self.rig.select_set(True)
    
        # ボーンの更新
        for bone_id in range(MocopiData.MAX_BONE):
            self.__bone_update(self.rig, bone_id, self.prop.is_mirror)

        # アクション
        self.__action_update()

        self.updating = False

    def has_rig(self) -> bool:
        if not utils.is_valid(self.rig):
            return False
        return self.rig.type == 'ARMATURE'

    def set_rig(self, rig: bpy.types.Object):
        self.rig = rig
        self.rig_name = rig.name if utils.is_valid(rig) else self.rig_name

    def retarget(self):

        if not utils.is_valid(self.rig) or self.rig.type != 'ARMATURE' or not utils.is_valid(self.prop):
            return
        
        self.bone_data_list[0].bone = self.rig.pose.bones.get(self.prop.root)
        self.bone_data_list[1].bone = self.rig.pose.bones.get(self.prop.torso_1)
        self.bone_data_list[2].bone = self.rig.pose.bones.get(self.prop.torso_2)
        self.bone_data_list[3].bone = self.rig.pose.bones.get(self.prop.torso_3)
        self.bone_data_list[4].bone = self.rig.pose.bones.get(self.prop.torso_4)
        self.bone_data_list[5].bone = self.rig.pose.bones.get(self.prop.torso_5)
        self.bone_data_list[6].bone = self.rig.pose.bones.get(self.prop.torso_6)
        self.bone_data_list[7].bone = self.rig.pose.bones.get(self.prop.torso_7)
        self.bone_data_list[8].bone = self.rig.pose.bones.get(self.prop.neck_1)
        self.bone_data_list[9].bone = self.rig.pose.bones.get(self.prop.neck_2)
        self.bone_data_list[10].bone = self.rig.pose.bones.get(self.prop.head)
        self.bone_data_list[11].bone = self.rig.pose.bones.get(self.prop.l_shoulder)
        self.bone_data_list[12].bone = self.rig.pose.bones.get(self.prop.l_up_arm)
        self.bone_data_list[13].bone = self.rig.pose.bones.get(self.prop.l_low_arm)
        self.bone_data_list[14].bone = self.rig.pose.bones.get(self.prop.l_hand)
        self.bone_data_list[15].bone = self.rig.pose.bones.get(self.prop.r_shoulder)
        self.bone_data_list[16].bone = self.rig.pose.bones.get(self.prop.r_up_arm)
        self.bone_data_list[17].bone = self.rig.pose.bones.get(self.prop.r_low_arm)
        self.bone_data_list[18].bone = self.rig.pose.bones.get(self.prop.r_hand)
        self.bone_data_list[19].bone = self.rig.pose.bones.get(self.prop.l_up_leg)
        self.bone_data_list[20].bone = self.rig.pose.bones.get(self.prop.l_low_leg)
        self.bone_data_list[21].bone = self.rig.pose.bones.get(self.prop.l_foot)
        self.bone_data_list[22].bone = self.rig.pose.bones.get(self.prop.l_toe)
        self.bone_data_list[23].bone = self.rig.pose.bones.get(self.prop.r_up_leg)
        self.bone_data_list[24].bone = self.rig.pose.bones.get(self.prop.r_low_leg)
        self.bone_data_list[25].bone = self.rig.pose.bones.get(self.prop.r_foot)
        self.bone_data_list[26].bone = self.rig.pose.bones.get(self.prop.r_toe)

    def run(self):

        if not utils.is_valid(self.rig):
            return

        self.running = True
        self.location0 = None

        asyncio.run(self.__run())

        # スケール計算
        self.rig_scale = self.__get_scale_armature(self.rig) if self.has_rig() else 1.0

        # アクション
        if not self.rig.animation_data or not self.rig.animation_data.action:
            self.rig.animation_data_create()
            self.rig.animation_data.action = bpy.data.actions.new(name=f"mocopi_action_{self.id}")
        self.action = self.rig.animation_data.action
        
    def stop(self):

        self.running = False

        if self.udp_socket:
            self.udp_socket.close()
        self.udp_socket = None

        if self.thread:
            self.thread.join()
        self.thread = None

    # Private

    async def __run(self):

        try:
            self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.udp_socket.bind(('', self.prop.port))
            self.udp_socket.settimeout(5)

            self.thread = threading.Thread(target=self.__listen)
            self.thread.daemon = True
            self.thread.start()
        except Exception as e:
            self.stop()

    def __listen(self):
        while self.running:
            try:
                data, _ = self.udp_socket.recvfrom(4096) # 1024, 4096
                if data and not self.updating:
                    self.last_data = MocopiData(data)
            except socket.timeout:
                print("Error occurred: socket.timeout")
            except Exception as e:
                print(f"Error occurred: {e}")
    
    def __bone_update(self, rig: bpy.types.Object, bone_id: int, is_mirror: bool):

        if not self.has_rig():
            return
        
        tran = self.last_data.get_tran(bone_id)
        if not tran:
            return

        bone = self.__get_bone_data(bone_id, is_mirror).bone
        if not bone:
            return

        # 位置
        px = -tran[4]
        py =  tran[5]
        pz =  tran[6]

        # 回転
        qx = -tran[0]
        qy =  tran[1]
        qz =  tran[2]
        qw = -tran[3]

        # Root
        if bone_id == 0: # Root

            # ミラー
            if is_mirror:
                px = -px

            location = self.__get_position(px, py, pz, 4) # 4
            location *= self.rig_scale / utils.get_scale_value(self.rig)

            if not self.location0:
                self.location0 = location
            # rig.location = location - self.location0
            bone.location = location - self.location0
            
            # rig.rotation_mode = 'QUATERNION'
            # rig.rotation_quaternion = self.__get_rotation(qw, qx, qy, qz, 19) # 28, 18, 19
            # return
        
        # ボーンの回転
        bone.rotation_mode = 'QUATERNION'

        if 11 <= bone_id and bone_id <= 14:

            # ミラー
            if is_mirror:
                qz = -qz
                qw = -qw
            
            # 左腕
            bone.rotation_quaternion = self.__get_rotation(qw, qx, qy, qz, 77)
            pass
        
        elif 15 <= bone_id and bone_id <= 18:

            # ミラー
            if is_mirror:
                qz = -qz
                qw = -qw

            # 右腕
            bone.rotation_quaternion = self.__get_rotation(qw, qx, qy, qz, 70)
            pass

        elif 19 <= bone_id and bone_id <= 22:

            # ミラー
            if is_mirror:
                qx = -qx
                qw = -qw

            # 左足
            bone.rotation_quaternion = self.__get_rotation(qw, qx, qy, qz, 15)
            pass

        elif 23 <= bone_id and bone_id <= 26:

            # ミラー
            if is_mirror:
                qx = -qx
                qw = -qw

            # 右足
            bone.rotation_quaternion = self.__get_rotation(qw, qx, qy, qz, 0) # 13 80 92
            pass

        elif 8 <= bone_id and bone_id <= 10:

            # ミラー
            if is_mirror:
                qx = -qx
                qw = -qw

            # 首・顔
            bone.rotation_quaternion = self.__get_rotation(qw, qx, qy, qz, 3) # 3 12
            pass

        elif bone_id == 0:

            # ミラー
            if is_mirror:
                qx = -qx
                qw = -qw

            # Root
            bone.rotation_quaternion = self.__get_rotation(qw, qx, qy, qz, 3) # 3 12
            pass

        else:

            # ミラー
            if is_mirror:
                qx = -qx
                qw = -qw

            # その他
            bone.rotation_quaternion = self.__get_rotation(qw, qx, qy, qz, 2) # 2 16 0 12, 87, 18, 21
            pass

    def __action_update(self):

        if not utils.is_valid(self.action):
            return
        
        scene = bpy.context.scene
        fcurves = self.action.fcurves

        if utils.is_valid(self.rig):
            for bone_id in range(MocopiData.MAX_BONE):
                bone = self.bone_data_list[bone_id].bone
                if bone:
                    if bone_id == 0:
                        for i in range(3): # x, y, z
                            data_path = f'pose.bones["{bone.name}"].location'
                            fcurve = fcurves.find(data_path, index=i)
                            if not fcurve:
                                fcurve = fcurves.new(data_path, index=i)
                            fcurve.keyframe_points.insert(scene.frame_current, bone.location[i])

                    for i in range(4): # x, y, z, w
                        data_path = f'pose.bones["{bone.name}"].rotation_quaternion'
                        fcurve = fcurves.find(data_path, index=i)
                        if not fcurve:
                            fcurve = fcurves.new(data_path, index=i)
                        fcurve.keyframe_points.insert(scene.frame_current, bone.rotation_quaternion[i])

        # # 繰り返し
        # if scene.frame_current > scene.frame_end:
        #     scene.frame_current = scene.frame_start

    def __get_scale_armature(self, rig: bpy.types.Object) -> float:
        
        bone_min = None
        bone_min_root = None

        if self.bone_data_list[0].bone is None:
            return 1.0

        for data in self.bone_data_list:

            if not data.bone:
                continue

            bone_z = (rig.matrix_world @ data.bone.head)[2]
            if data.bone.name == self.bone_data_list[0].bone.name:
                if bone_min_root is None or bone_min_root > bone_z:
                    bone_min_root = bone_z
            if bone_min is None or bone_min > bone_z:
                bone_min = bone_z

        if not bone_min_root or not bone_min:
            return 1.0
        
        height = bone_min_root - bone_min
        if height == 0:
            return 1.0

        scale_factor = height / 0.899717666208744
        return scale_factor
    
    def __get_bone_data(self, bone_id: int, is_mirror: bool) -> BoneData:

        mirror_map = {
            11: 15, 12: 16, 13: 17, 14: 18,  # 左腕 -> 右腕
            15: 11, 16: 12, 17: 13, 18: 14,  # 右腕 -> 左腕
            19: 23, 20: 24, 21: 25, 22: 26,  # 左足 -> 右足
            23: 19, 24: 20, 25: 21, 26: 22   # 右足 -> 左足
        }

        if is_mirror and bone_id in mirror_map:
            bone_id = mirror_map[bone_id]

        return self.bone_data_list[bone_id]
    
    def __get_position(self, px: float, py: float, pz: float, id: int) -> Vector:
        position = Vector((px, py, pz))
        if id == 0:
            position = Vector((px, py, pz))
        elif id == 1:
            position = Vector((px, -py, pz))
        elif id == 2:
            position = Vector((px, py, -pz))
        elif id == 3:
            position = Vector((px, -py, -pz))
        elif id == 4:
            position = Vector((-px, py, pz))
        elif id == 5:
            position = Vector((-px, -py, pz))
        elif id == 6:
            position = Vector((-px, py, -pz))
        elif id == 7:
            position = Vector((-px, -py, -pz))

        elif id == 8:
            position = Vector((px, pz, py))
        elif id == 9:
            position = Vector((px, -pz, py))
        elif id == 10:
            position = Vector((px, pz, -py))
        elif id == 11:
            position = Vector((px, -pz, -py))
        elif id == 12:
            position = Vector((-px, pz, py))
        elif id == 13:
            position = Vector((-px, -pz, py))
        elif id == 14:
            position = Vector((-px, pz, -py))
        elif id == 15:
            position = Vector((-px, -pz, -py))

        elif id == 16:
            position = Vector((py, px, pz))
        elif id == 17:
            position = Vector((py, -px, pz))
        elif id == 18:
            position = Vector((py, px, -pz))
        elif id == 19:
            position = Vector((py, -px, -pz))
        elif id == 20:
            position = Vector((-py, px, pz))
        elif id == 21:
            position = Vector((-py, -px, pz))
        elif id == 22:
            position = Vector((-py, px, -pz))
        elif id == 23:
            position = Vector((-py, -px, -pz))

        elif id == 24:
            position = Vector((py, pz, px))
        elif id == 25:
            position = Vector((py, -pz, px))
        elif id == 26:
            position = Vector((py, pz, -px))
        elif id == 27:
            position = Vector((py, -pz, -px))
        elif id == 28:
            position = Vector((-py, pz, px))
        elif id == 29:
            position = Vector((-py, -pz, px))
        elif id == 30:
            position = Vector((-py, pz, -px))
        elif id == 31:
            position = Vector((-py, -pz, -px))

        elif id == 32:
            position = Vector((pz, px, py))
        elif id == 33:
            position = Vector((pz, -px, py))
        elif id == 34:
            position = Vector((pz, px, -py))
        elif id == 35:
            position = Vector((pz, -px, -py))
        elif id == 36:
            position = Vector((-pz, px, py))
        elif id == 37:
            position = Vector((-pz, -px, py))
        elif id == 38:
            position = Vector((-pz, px, -py))
        elif id == 39:
            position = Vector((-pz, -px, -py))

        elif id == 40:
            position = Vector((pz, py, px))
        elif id == 41:
            position = Vector((pz, -py, px))
        elif id == 42:
            position = Vector((pz, py, -px))
        elif id == 43:
            position = Vector((pz, -py, -px))
        elif id == 44:
            position = Vector((-pz, py, px))
        elif id == 45:
            position = Vector((-pz, -py, px))
        elif id == 46:
            position = Vector((-pz, py, -px))
        elif id == 47:
            position = Vector((-pz, -py, -px))

        return position
    
    def __get_rotation(self, qw: float, qx: float, qy: float, qz: float, id: int) -> Quaternion:
        rotation = Quaternion([qw, qx, qy, qz])
        if (id == 0):
            rotation = Quaternion([qw, qx, qy, qz])
        elif (id == 1):
            rotation = Quaternion([qw, qx, -qy, qz])
        elif (id == 2):
            rotation = Quaternion([qw, qx, qy, -qz])
        elif (id == 3):
            rotation = Quaternion([qw, qx, -qy, -qz])
        elif (id == 4):
            rotation = Quaternion([qw, -qx, qy, qz])
        elif (id == 5):
            rotation = Quaternion([qw, -qx, -qy, qz])
        elif (id == 6):
            rotation = Quaternion([qw, -qx, qy, -qz])
        elif (id == 7):
            rotation = Quaternion([qw, -qx, -qy, -qz])

        elif (id == 8):
            rotation = Quaternion([-qw, qx, qy, qz])
        elif (id == 9):
            rotation = Quaternion([-qw, qx, -qy, qz])
        elif (id == 10):
            rotation = Quaternion([-qw, qx, qy, -qz])
        elif (id == 11):
            rotation = Quaternion([-qw, qx, -qy, -qz])
        elif (id == 12):
            rotation = Quaternion([-qw, -qx, qy, qz])
        elif (id == 13):
            rotation = Quaternion([-qw, -qx, -qy, qz])
        elif (id == 14):
            rotation = Quaternion([-qw, -qx, qy, -qz])
        elif (id == 15):
            rotation = Quaternion([-qw, -qx, -qy, -qz])

        elif (id == 16):
            rotation = Quaternion([qw, qx, qz, qy])
        elif (id == 17):
            rotation = Quaternion([qw, qx, -qz, qy])
        elif (id == 18):
            rotation = Quaternion([qw, qx, qz, -qy])
        elif (id == 19):
            rotation = Quaternion([qw, qx, -qz, -qy])
        elif (id == 20):
            rotation = Quaternion([qw, -qx, qz, qy])
        elif (id == 21):
            rotation = Quaternion([qw, -qx, -qz, qy])
        elif (id == 22):
            rotation = Quaternion([qw, -qx, qz, -qy])
        elif (id == 23):
            rotation = Quaternion([qw, -qx, -qz, -qy])
        
        elif (id == 24):
            rotation = Quaternion([-qw, qx, qz, qy])
        elif (id == 25):
            rotation = Quaternion([-qw, qx, -qz, qy])
        elif (id == 26):
            rotation = Quaternion([-qw, qx, qz, -qy])
        elif (id == 27):
            rotation = Quaternion([-qw, qx, -qz, -qy])
        elif (id == 28):
            rotation = Quaternion([-qw, -qx, qz, qy])
        elif (id == 29):
            rotation = Quaternion([-qw, -qx, -qz, qy])
        elif (id == 30):
            rotation = Quaternion([-qw, -qx, qz, -qy])
        elif (id == 31):
            rotation = Quaternion([-qw, -qx, -qz, -qy])

        if (id == 32):
            rotation = Quaternion([qw, qy, qx, qz])
        elif (id == 33):
            rotation = Quaternion([qw, -qy, qx, qz])
        elif (id == 34):
            rotation = Quaternion([qw, -qy, -qx, qz])
        elif (id == 35):
            rotation = Quaternion([qw, -qy, qx, -qz])
        elif (id == 36):
            rotation = Quaternion([qw, -qy, -qx, -qz])
        elif (id == 37):
            rotation = Quaternion([qw, qy, -qx, qz])
        elif (id == 38):
            rotation = Quaternion([qw, qy, -qx, -qz])
        elif (id == 39):
            rotation = Quaternion([qw, qy, qx, -qz])

        elif (id == 40):
            rotation = Quaternion([-qw, qy, qx, qz])
        elif (id == 41):
            rotation = Quaternion([-qw, -qy, qx, qz])
        elif (id == 42):
            rotation = Quaternion([-qw, -qy, -qx, qz])
        elif (id == 43):
            rotation = Quaternion([-qw, -qy, qx, -qz])
        elif (id == 44):
            rotation = Quaternion([-qw, -qy, -qx, -qz])
        elif (id == 45):
            rotation = Quaternion([-qw, qy, -qx, qz])
        elif (id == 46):
            rotation = Quaternion([-qw, qy, -qx, -qz])
        elif (id == 47):
            rotation = Quaternion([-qw, qy, qx, -qz])

        elif (id == 48):
            rotation = Quaternion([qw, qy, qz, qx])
        elif (id == 49):
            rotation = Quaternion([qw, -qy, qz, qx])
        elif (id == 50):
            rotation = Quaternion([qw, -qy, -qz, qx])
        elif (id == 51):
            rotation = Quaternion([qw, -qy, qz, -qx])
        elif (id == 52):
            rotation = Quaternion([qw, -qy, -qz, -qx])
        elif (id == 53):
            rotation = Quaternion([qw, qy, -qz, qx])
        elif (id == 54):
            rotation = Quaternion([qw, qy, -qz, -qx])
        elif (id == 55):
            rotation = Quaternion([qw, qy, qz, -qx])

        elif (id == 56):
            rotation = Quaternion([-qw, qy, qz, qx])
        elif (id == 57):
            rotation = Quaternion([-qw, -qy, qz, qx])
        elif (id == 58):
            rotation = Quaternion([-qw, -qy, -qz, qx])
        elif (id == 59):
            rotation = Quaternion([-qw, -qy, qz, -qx])
        elif (id == 60):
            rotation = Quaternion([-qw, -qy, -qz, -qx])
        elif (id == 61):
            rotation = Quaternion([-qw, qy, -qz, qx])
        elif (id == 62):
            rotation = Quaternion([-qw, qy, -qz, -qx])
        elif (id == 63):
            rotation = Quaternion([-qw, qy, qz, -qx])

        elif (id == 64):
            rotation = Quaternion([qw, qz, qx, qy])
        elif (id == 65):
            rotation = Quaternion([qw, -qz, qx, qy])
        elif (id == 66):
            rotation = Quaternion([qw, -qz, -qx, qy])
        elif (id == 67):
            rotation = Quaternion([qw, -qz, qx, -qy])
        elif (id == 68):
            rotation = Quaternion([qw, -qz, -qx, -qy])
        elif (id == 69):
            rotation = Quaternion([qw, qz, -qx, qy])
        elif (id == 70):
            rotation = Quaternion([qw, qz, -qx, -qy])
        elif (id == 71):
            rotation = Quaternion([qw, qz, qx, -qy])

        elif (id == 72):
            rotation = Quaternion([-qw, qz, qx, qy])
        elif (id == 73):
            rotation = Quaternion([-qw, -qz, qx, qy])
        elif (id == 74):
            rotation = Quaternion([-qw, -qz, -qx, qy])
        elif (id == 75):
            rotation = Quaternion([-qw, -qz, qx, -qy])
        elif (id == 76):
            rotation = Quaternion([-qw, -qz, -qx, -qy])
        elif (id == 77):
            rotation = Quaternion([-qw, qz, -qx, qy])
        elif (id == 78):
            rotation = Quaternion([-qw, qz, -qx, -qy])
        elif (id == 79):
            rotation = Quaternion([-qw, qz, qx, -qy])

        elif (id == 80):
            rotation = Quaternion([qw, qz, qy, qx])
        elif (id == 81):
            rotation = Quaternion([qw, -qz, qy, qx])
        elif (id == 82):
            rotation = Quaternion([qw, -qz, -qy, qx])
        elif (id == 83):
            rotation = Quaternion([qw, -qz, qy, -qx])
        elif (id == 84):
            rotation = Quaternion([qw, -qz, -qy, -qx])
        elif (id == 85):
            rotation = Quaternion([qw, qz, -qy, qx])
        elif (id == 86):
            rotation = Quaternion([qw, qz, -qy, -qx])
        elif (id == 87):
            rotation = Quaternion([qw, qz, qy, -qx])

        elif (id == 88):
            rotation = Quaternion([-qw, qz, qy, qx])
        elif (id == 89):
            rotation = Quaternion([-qw, -qz, qy, qx])
        elif (id == 90):
            rotation = Quaternion([-qw, -qz, -qy, qx])
        elif (id == 91):
            rotation = Quaternion([-qw, -qz, qy, -qx])
        elif (id == 92):
            rotation = Quaternion([-qw, -qz, -qy, -qx])
        elif (id == 93):
            rotation = Quaternion([-qw, qz, -qy, qx])
        elif (id == 94):
            rotation = Quaternion([-qw, qz, -qy, -qx])
        elif (id == 95):
            rotation = Quaternion([-qw, qz, qy, -qx])

        return rotation

    def __mat3_to_vec_roll(self, mat):
        vecmat = self.__vec_roll_to_mat3(mat.col[1], 0)
        vecmatinv = vecmat.inverted()
        rollmat = vecmatinv @ mat
        roll = math.atan2(rollmat[0][2], rollmat[2][2])
        return roll
    
    def __vec_roll_to_mat3(self, vec, roll):
        target = Vector((0, 0.1, 0))
        nor = vec.normalized()
        axis = target.cross(nor)
        if axis.dot(axis) > 0.0000000001:
            axis.normalize()
            theta = target.angle(nor)
            bMatrix = Matrix.Rotation(theta, 3, axis)
        else:
            updown = 1 if target.dot(nor) > 0 else -1
            bMatrix = Matrix.Scale(updown, 3)
            bMatrix[2][2] = 1.0
        rMatrix = Matrix.Rotation(roll, 3, nor)
        mat = rMatrix @ bMatrix
        return mat


class MocopiData:

    MAX_BONE = 27

    def __init__(self, data):

        self.head = {}
        self.sndf = {}
        self.fram = {}
        self.btrs = {}

        self.__init(data)

    def get_tran(self, bone_id: int) -> list[float]:
        # return self.btrs['btdt'][bone_id]['tran']['data']

        btdt = next((btdt for btdt in self.btrs['btdt'] if btdt['bnid']['data'] == bone_id), None)
        return btdt['tran']['data'] if btdt else None

    def __init(self, data):
        
        i = 0

        # head
        head = self.__parse(data, i)
        i = head['end']

        # head > ftyp
        ftyp = self.__parse(data, i, 'ascii')
        i = ftyp['end']

        # head > vrsn
        vrsn = self.__parse(data, i, 'int')
        i = vrsn['end']

        self.head = head
        self.head['ftyp'] = ftyp
        self.head['vrsn'] = vrsn

        # sndf
        sndf = self.__parse(data, i)
        i = sndf['end']

        # sndf > ipad
        ipad = self.__parse(data, i, 'raw')
        i = ipad['end']

        # sndf > rcvp
        rcvp = self.__parse(data, i, 'raw')
        i = rcvp['end']

        self.sndf = sndf
        self.sndf['ipad'] = ipad
        self.sndf['rcvp'] = rcvp

        # fram
        fram = self.__parse(data, i)
        if fram['name'] == 'fram':
            i = fram['end']
            self.fram = fram

        # fram > fnum
        fnum = self.__parse(data, i, 'int')
        if fnum['name'] == 'fnum':
            i = fnum['end']
            self.fram['fnum'] = fnum

        # fram > time
        time = self.__parse(data, i, 'int')
        if time['name'] == 'time':
            i = time['end']
            self.fram['time'] = time

        # fram > uttm
        uttm = self.__parse(data, i, 'int')
        if uttm['name'] == 'uttm':
            i = uttm['end']
            self.fram['uttm'] = uttm

        # # fram > tmcd
        tmcd = self.__parse(data, i, 'int')
        if tmcd['name'] == 'tmcd':
            i = tmcd['end']
            self.fram['tmcd'] = tmcd

        # btrs
        btrs = self.__parse(data, i)
        i = btrs['end']

        # btrs > btdt
        btdt_list = []
        for _ in range(MocopiData.MAX_BONE):
            bone = self.__btdt_parse(data, i)
            i = bone['end']
            btdt_list.append(bone)

        self.btrs = btrs
        self.btrs['btdt'] = btdt_list
        
    def __parse(self, bytes: bytes, i: int, type: str = None) -> dict:

        SIZE = 4

        size = int.from_bytes(bytes[i:i+SIZE], 'little')
        i += SIZE
        name = bytes[i:i+SIZE].decode('ascii')
        i += SIZE

        data = None
        if type == 'ascii':
            data = bytes[i:i+size].decode('ascii')
            i += size
        elif type == 'int':
            data = int.from_bytes(bytes[i:i+size], 'little')
            i += size
        elif type == 'vector':
            data = struct.unpack('<7f', bytes[i:i+size])
            i += size
        elif type == 'raw':
            data = bytes[i:i+size]
            i += size
        
        return {
                'size': size, 
                'name': name, 
                'data': data, 
                'end': i
            }
    
    def __btdt_parse(self, bytes: bytes, i: int) -> dict:

        # btdt
        btdt = self.__parse(bytes, i)
        i = btdt['end']

        # btdt > bnid
        bnid = self.__parse(bytes, i, 'int')
        i = bnid['end']

        # btdt > tran
        tran = self.__parse(bytes, i, 'vector')
        i = tran['end']

        btdt['bnid'] = bnid
        btdt['tran'] = tran
        btdt['end'] = i

        return btdt
